; IDL interface for Zhang's MS-VST executables
;---------------------------------
; anscfiszatrous2D 	-- Anscombe and Fisz VST with the a trous B3-spline WT.
;
; msvstudwt2D	 	-- MSVST with the 2D bi-orthogonal undecimeted WT (7/9).
;
; msvstatrouscoupled2D 	-- MSVST with the a trous B3-spline WT with two different filter-banks.
;			   The direct reconstruction (possible here thansk to the particular WT) 
;			   is used as an initialization for the iterative reconstruction.
;
; msvstatrousseparated2D 		-- Same as msvstatrouscoupled2D but different initialization (separated mode).
;
; msvstatrousseparated2D1D	-- Multichannel data MSVST-based denoising.
;
; msvstatrouscorrectedcoupled2D 	-- Same as msvstatrouscoupled2D but the initial estimate is corrected for bias introduced by inverting the MSVST.
;
; See the papers for details.
;---------------------------------
;
;INPUT
;    
;    data : name of the fits file containing the data to denoise (data.fits)
;    (optional) model : namle of the fits file containing the background model to extract (model.fits)
;
;OUTPUT
;
;    result : name of the fits file containing the denoised data (result.fits)
;
;PARAMETERS
;
;    FDR : use FDR-based denoising threshold (0 or 1) (default = 0)
;    nScale : number of scales used (range: [1, 7]) (default = 5)
;    fScale : coefficients at any scale smaller than fScale are treated insignificant (range: [1, nScale]) (default = 1)
;    nIter : number of iterations (> 1) (default = 10)
;    AnscFisz : only for anscfiszatrous2D : if 0, use Anscombe (default), if 1, use Fisz
;
;EXAMPLE
;
; Compute the MSVST with the a trous B3-spline WT in coupled mode on data file data.fits
; Result is stocked in file result.fits
;            msvstatrouscoupled2D,data,result
;
; Author : J. Schmitt (2009)


pro msvstatrouscoupled2D,data,result,model=model,FDR=FDR,nScale=nScale,fScale=fScale,nIter=nIter

if not keyword_set(FDR) then FDR = 0
if not keyword_set(nScale) then nScale=5
if not keyword_set(fScale) then fScale=1
if not keyword_set(nIter) then nIter=10

if keyword_set(model) then cmd='./msvstatrouscoupled2D -M '+FDR+' -n '+nScale+' -F '+fScale+' -i '+niter+' '+data+'.fits '+result+'.fits '+model+'.fits'
else cmd='./msvstatrouscoupled2D -M '+FDR+' -n '+nScale+' -F '+fScale+' -i '+niter+' '+data+'.fits '+result+'.fits'
spawn,cmd

end

;-----------------------------------------------------------------------------------------------------------

pro msvstatrouscorrectedcoupled2D,data,result,model=model,FDR=FDR,nScale=nScale,fScale=fScale,nIter=nIter

if not keyword_set(FDR) then FDR = 0
if not keyword_set(nScale) then nScale=5
if not keyword_set(fScale) then fScale=1
if not keyword_set(nIter) then nIter=10

if keyword_set(model) cmd='./msvstatrouscorrectedcoupled2D -M '+FDR+' -n '+nScale+' -F '+fScale+' -i '+niter+' '+data+'.fits '+result+'.fits '+model+'.fits'
else cmd='./msvstatrouscorrectedcoupled2D -M '+FDR+' -n '+nScale+' -F '+fScale+' -i '+niter+' '+data+'.fits '+result+'.fits'
spawn,cmd

end

;-----------------------------------------------------------------------------------------------------------

pro msvstatrousseparated2D,data,result,FDR=FDR,nScale=nScale,fScale=fScale,nIter=nIter

if not keyword_set(FDR) then FDR = 0
if not keyword_set(nScale) then nScale=5
if not keyword_set(fScale) then fScale=1
if not keyword_set(nIter) then nIter=10

cmd='./msvstatrousseparated2D -M '+FDR+' -n '+nScale+' -F '+fScale+' -i '+niter+' '+data+'.fits '+result+'.fits'
spawn,cmd

end

;-----------------------------------------------------------------------------------------------------------

pro msvstatrousseparated2D1D,data,result,model=model,FDR=FDR,nScale=nScale,fScale=fScale,nIter=nIter

if not keyword_set(FDR) then FDR = 0
if not keyword_set(nScale) then nScale=5
if not keyword_set(fScale) then fScale=1
if not keyword_set(nIter) then nIter=10

if keyword_set(model) cmd='./msvstatrousseparated2D1D -M '+FDR+' -n '+nScale+' -F '+fScale+' -i '+niter+' '+data+'.fits '+result+'.fits '+model+'.fits'
else cmd='./msvstatrousseparated2D1D -M '+FDR+' -n '+nScale+' -F '+fScale+' -i '+niter+' '+data+'.fits '+result+'.fits'
spawn,cmd

end

;-----------------------------------------------------------------------------------------------------------

pro msvstatudwt2D,data,result,FDR=FDR,nScale=nScale,fScale=fScale,nIter=nIter

if not keyword_set(FDR) then FDR = 0
if not keyword_set(nScale) then nScale=5
if not keyword_set(fScale) then fScale=1
if not keyword_set(nIter) then nIter=10

cmd='./msvstatudwt2D -M '+FDR+' -n '+nScale+' -F '+fScale+' -i '+niter+' '+data+'.fits '+result+'.fits'
spawn,cmd

end

;-----------------------------------------------------------------------------------------------------------

pro anscfiszatrous2D,data,result,AnscFisz=AnscFisz,FDR=FDR,nScale=nScale,fScale=fScale,nIter=nIter

if not keyword_set(AnscFisz) then AnscFisz=0
if not keyword_set(FDR) then FDR = 0
if not keyword_set(nScale) then nScale=5
if not keyword_set(fScale) then fScale=1
if not keyword_set(nIter) then nIter=0

cmd='./anscfiszatrous2D -M '+AnscFisz+' -m '+FDR+' -n '+nScale+' -F '+fScale+' -i '+niter+' '+data+'.fits '+result+'.fits'
spawn,cmd

end